@@ -1409,59 +1409,66 @@ void Smb4KMounter::checkAccessibility( Smb4KShare *share )
 
 void Smb4KMounter::timerEvent( QTimerEvent * )
 {
-  if ( !m_working && !m_queue.isEmpty() )
+  if ( !kapp->startingUp() )
   {
-    // Tell the mounter, that it is busy.
-    m_working = true;
-
-    QString *item = m_queue.dequeue();
-    int todo = item->section( ":", 0, 0 ).toInt();
-
-    switch ( todo )
+    if ( !m_working && !m_queue.isEmpty() )
     {
-      case Remount:
+      // Tell the mounter, that it is busy.
+      m_working = true;
+
+      QString *item = m_queue.dequeue();
+      int todo = item->section( ":", 0, 0 ).toInt();
+
+      switch ( todo )
       {
-        remount();
-        break;
-      }
-      case Import:
-      {
-        import();
-        break;
-      }
-      case Mount:
-      {
-        emit state( MOUNTER_MOUNTING );
-        mount( item->section( ":", 1, 1 ), item->section( ":", 2, 2 ), item->section( ":", 3, 3 ), item->section( ":", 4, 4 ) );
-        break;
-      }
-      case Unmount:
-      {
-        emit state( MOUNTER_UNMOUNTING );
-        unmount( item->section( ":", 1, 1 ), (bool)item->section( ":", 2, 2 ).toInt() /* force */, (bool)item->section( ":", 3, 3 ).toInt() /* noMessage */);
-        break;
-      }
-      case UnmountAll:
-      {
-        unmountAll();
-        break;
-      }
-      default:
-      {
-        break;
+        case Remount:
+        {
+          remount();
+          break;
+        }
+        case Import:
+        {
+          import();
+          break;
+        }
+        case Mount:
+        {
+          emit state( MOUNTER_MOUNTING );
+          mount( item->section( ":", 1, 1 ), item->section( ":", 2, 2 ), item->section( ":", 3, 3 ), item->section( ":", 4, 4 ) );
+          break;
+        }
+        case Unmount:
+        {
+          emit state( MOUNTER_UNMOUNTING );
+          unmount( item->section( ":", 1, 1 ), (bool)item->section( ":", 2, 2 ).toInt() /* force */, (bool)item->section( ":", 3, 3 ).toInt() /* noMessage */);
+          break;
+        }
+        case UnmountAll:
+        {
+          unmountAll();
+          break;
+        }
+        default:
+        {
+          break;
+        }
       }
+
+      delete item;
     }
 
-    delete item;
+    m_priv->timerTicks++;
+
+    if ( m_priv->timerTicks * timerInterval() >= Smb4KSettings::checkInterval() /* msec */ &&
+        (!m_working || m_queue.isEmpty()) )
+    {
+      m_queue.enqueue( new QString( QString( "%1:" ).arg( Import ) ) );
+      m_priv->timerTicks = 0;
+    }
   }
-
-  m_priv->timerTicks++;
-
-  if ( m_priv->timerTicks * timerInterval() >= Smb4KSettings::checkInterval() /* msec */ &&
-       (!m_working || m_queue.isEmpty()) )
+  else
   {
-    m_queue.enqueue( new QString( QString( "%1:" ).arg( Import ) ) );
-    m_priv->timerTicks = 0;
+    // Do nothing until the application has started up.
   }
 }
 
