@@ -42,6 +42,8 @@
 #include <klineedit.h>
 #include <kcombobox.h>
 #include <kdebug.h>
+#include <kapplication.h>
+#include <dcopclient.h>
 
 #ifdef __FreeBSD__
 #include <sys/stat.h>
@@ -59,10 +61,11 @@
 
 #ifndef __FreeBSD__
 Smb4KPasswordHandler::Smb4KPasswordHandler( Smb4KHomesSharesHandler *handler, QObject *parent, const char *name )
-: QObject( parent, name ), m_handler( handler )
+: QObject( parent, name ), m_handler( handler ), m_wallet_support_disabled( false )
 #else
 Smb4KPasswordHandler::Smb4KPasswordHandler( Smb4KHomesSharesHandler *s_handler, Smb4KSambaOptionsHandler *o_handler, QObject *parent, const char *name )
-: QObject( parent, name ), m_handler( s_handler ), m_options_handler( o_handler )
+: QObject( parent, name ), m_handler( s_handler ), m_options_handler( o_handler ),
+  m_wallet_support_disabled( false )
 #endif
 {
   if ( !m_handler )
@@ -99,10 +102,22 @@ Smb4KPasswordHandler::~Smb4KPasswordHandler()
 
 void Smb4KPasswordHandler::open_close_wallet()
 {
-  if ( Smb4KSettings::useWallet() )
+  if ( Smb4KSettings::useWallet() && !walletSupportIsDisabled() )
   {
-    if ( !m_wallet || (m_wallet && !m_wallet->isOpen()) )
+    if ( !walletIsOpen() )
     {
+      // Start the wallet manager before accessing the wallet. We
+      // do not care about the return value, because we do not do
+      // error handling here.
+      if ( kapp )
+      {
+        (void) kapp->kdeinitExec( "kwalletmanager" );
+      }
+      else
+      {
+        // Do nothing. --- Good luck!
+      }
+
       m_wallet = KWallet::Wallet::openWallet( KWallet::Wallet::NetworkWallet(), 0, KWallet::Wallet::Synchronous );
 
       if ( m_wallet )
@@ -110,9 +125,14 @@ void Smb4KPasswordHandler::open_close_wallet()
         if ( !m_wallet->hasFolder( "Smb4K" ) )
         {
           m_wallet->createFolder( "Smb4K" );
+          m_wallet->setFolder( "Smb4K" );
         }
+        else
+        {
+          m_wallet->setFolder( "Smb4K" );
 
-        m_wallet->setFolder( "Smb4K" );
+          convert_old_entries();
+        }
       }
       else
       {
@@ -121,10 +141,13 @@ void Smb4KPasswordHandler::open_close_wallet()
         delete m_wallet;
         m_wallet = NULL;
 
-        Smb4KSettings::setUseWallet( false );
-        Smb4KSettings::writeConfig();
+        m_wallet_support_disabled = true;
       }
     }
+    else
+    {
+      convert_old_entries();
+    }
   }
   else
   {
@@ -137,6 +160,74 @@ void Smb4KPasswordHandler::open_close_wallet()
 }
 
 
+void Smb4KPasswordHandler::convert_old_entries()
+{
+  // Convert old entries (Smb4K 0.9.2 and earlier) to the
+  // new map based format.
+  if ( !m_wallet->entryList().isEmpty() )
+  {
+    QStringList entries = m_wallet->entryList();
+
+    // Since in the old format the keys contained a string like
+    // this: HOST:USER, we only need to test whether the first
+    // key contains a ":". If that's the case, we need to convert
+    // the entries, otherwise we don't.
+    if ( entries.first().contains( ":" ) != 0 )
+    {
+      for ( QStringList::Iterator it = entries.begin(); it != entries.end(); ++it )
+      {
+        // Get the password.
+        QString pass;
+        m_wallet->readPassword( *it, pass );
+
+        if ( (*it).startsWith( "DEFAULT:" ) )
+        {
+          // Convert the data to the new format.
+          QMap<QString,QString> map;
+          map["Login"]     = (*it).section( ":", 1, 1 );
+          map["Password"]  = pass;
+
+          m_wallet->writeMap( "DEFAULT_LOGIN", map );
+        }
+        else
+        {
+          // Convert the data to the new format.
+          QMap<QString,QString> map;
+          map["Login"]     = (*it).section( ":", 3, 3 );
+          map["Password"]  = pass;
+
+          if ( QString::compare( (*it).section( ":", 0, 0 ), "*" ) != 0 )
+          {
+            map["Workgroup"] = (*it).section( ":", 0, 0 ).upper();
+          }
+
+          if ( QString::compare( (*it).section( ":", 2, 2 ), "*" ) == 0 )
+          {
+            m_wallet->writeMap( (*it).section( ":", 1, 1 ).upper(), map );
+          }
+          else
+          {
+            m_wallet->writeMap( "//"+(*it).section( ":", 1, 1 ).upper()+
+                                "/"+(*it).section( ":", 2, 2 ).upper(), map );
+          }
+        }
+
+        // Delete the old entry.
+        m_wallet->removeEntry( *it );
+      }
+    }
+    else
+    {
+      // Do nothing.
+    }
+  }
+  else
+  {
+    // Do nothing
+  }
+}
+
+
 bool Smb4KPasswordHandler::askpass( const QString &workgroup, const QString &host, const QString &share, int desc, QWidget *parent, const char *name )
 {
   // m_auth is NULL:
@@ -320,137 +411,226 @@ Smb4KAuthInfo *Smb4KPasswordHandler::readAuth( Smb4KAuthInfo *authInfo )
 
   open_close_wallet();
 
-  if ( m_wallet && m_wallet->isOpen() )
+  if ( walletIsOpen() )
   {
-    QStringList full_list = m_wallet->entryList();
-    QStringList host_list = full_list.grep( authInfo->host(), false );
+    // Get the authentication information from the wallet.
+    // Always prefer either the exact match or the information
+    // that was provided for the host.
+    QMap<QString,QString> map;
 
-    if ( !host_list.isEmpty() )
+    if ( !authInfo->share().isEmpty() )
     {
-      for ( QStringList::ConstIterator it = host_list.begin(); it != host_list.end(); ++it )
+      m_wallet->readMap( "//"+authInfo->host().upper()+"/"+authInfo->share().upper(), map );
+
+      if ( map.isEmpty() )
       {
-         if ( !authInfo->share().isEmpty() &&
-              (*it).upper().contains( ":"+authInfo->host().upper()+":"+authInfo->share().upper() ) != 0 )
+        m_wallet->readMap( authInfo->host().upper(), map );
+
+        if ( map.isEmpty() )
         {
-          QString user = (*it).section( ":", 3, 3 ).stripWhiteSpace();
-          QString pass;
+          if ( Smb4KSettings::useDefaultLogin() )
+          {
+            m_wallet->readMap( "DEFAULT_LOGIN", map );
 
-          m_wallet->readPassword( *it, pass );
-
-          authInfo->setUser( user );
-          authInfo->setPassword( pass );
-
-          break;
-        }
-        else if ( !authInfo->share().isEmpty() &&
-                  (*it).upper().contains( ":"+authInfo->host().upper()+":*" ) != 0 )
-        {
-          QString user = (*it).section( ":", 3, 3 ).stripWhiteSpace();
-          QString pass;
-
-          m_wallet->readPassword( *it, pass );
-
-          authInfo->setUser( user );
-          authInfo->setPassword( pass );
-
-          continue;
-        }
-        else if ( authInfo->share().isEmpty() &&
-                  (*it).upper().contains( ":"+authInfo->host().upper()+":" ) )
-        {
-          QString user = (*it).section( ":", 3, 3 ).stripWhiteSpace();
-          QString pass;
-
-          m_wallet->readPassword( *it, pass );
-
-          authInfo->setUser( user );
-          authInfo->setPassword( pass );
-
-          break;
-        }
-        else if ( authInfo->share().isEmpty() && (*it).upper().startsWith( "DEFAULT:" ) &&
-                  Smb4KSettings::useDefaultLogin() )
-        {
-          QString user = (*it).section( ":", 1, 1 ).stripWhiteSpace();
-          QString pass;
-
-          m_wallet->readPassword( *it, pass );
-
-          authInfo->setUser( user );
-          authInfo->setPassword( pass );
-
-          continue;
+            if ( map.isEmpty() )
+            {
+              return authInfo;
+            }
+            else
+            {
+              // We have authentication information.
+              // So, do nothing here.
+            }
+          }
+          else
+          {
+            return authInfo;
+          }
         }
         else
         {
-          continue;
+          // Check that the workgroup is correct.
+          if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
+          {
+            if ( QString::compare( map["Workgroup"].upper(), authInfo->workgroup().upper() ) != 0 )
+            {
+              return authInfo;
+            }
+            else
+            {
+              // Everything is OK. Do nothing.
+            }
+          }
+          else
+          {
+            // We cannot check if the workgroup is OK. So, just return
+            // the authentication information we just collected. If it
+            // should be wrong, the user will be asked to provide the
+            // correct data anyway.
+          }
+        }
+      }
+      else
+      {
+        // Check that the workgroup is correct.
+        if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
+        {
+          if ( QString::compare( map["Workgroup"].upper(), authInfo->workgroup().upper() ) != 0 )
+          {
+            return authInfo;
+          }
+          else
+          {
+            // Everything is OK. Do nothing.
+          }
+        }
+        else
+        {
+          // We cannot check if the workgroup is OK. So, just process
+          // the authentication information we just collected. If it
+          // should be wrong, the user will be asked to provide the
+          // correct data anyway.
         }
       }
     }
     else
     {
-      if ( Smb4KSettings::useDefaultLogin() )
+      m_wallet->readMap( authInfo->host().upper(), map );
+
+      if ( map.isEmpty() )
       {
-        QStringList default_list = full_list.grep( "DEFAULT", false );
-        QString user, pass;
-
-        for ( QStringList::ConstIterator it = default_list.begin(); it != default_list.end(); ++it )
+        if ( Smb4KSettings::useDefaultLogin() )
         {
-          if ( (*it).startsWith( "DEFAULT:" ) )
+          m_wallet->readMap( "DEFAULT_LOGIN", map );
+
+          if ( map.isEmpty() )
           {
-            user = (*it).section( ":", 1, 1 ).stripWhiteSpace();
-            m_wallet->readPassword( *it, pass );
-
-            authInfo->setUser( user );
-            authInfo->setPassword( pass );
-
-            break;
+            return authInfo;
           }
-          else
-          {
-            continue;
-          }
-        }
-      }
-    }
-  }
-  else
-  {
-    if ( Smb4KSettings::rememberPasswords() && !m_auth_list.isEmpty() )
-    {
-      for ( QValueList<Smb4KAuthInfo *>::Iterator it = m_auth_list.begin(); it != m_auth_list.end(); ++it )
-      {
-        if ( (QString::compare( (*it)->workgroup().upper(), authInfo->workgroup().upper() ) == 0 &&
-              QString::compare( (*it)->host().upper(), authInfo->host().upper() ) == 0 &&
-              QString::compare( (*it)->share().upper(), authInfo->share().upper() ) == 0) ||
-             (QString::compare( (*it)->workgroup().upper(), "*" ) == 0 &&
-              QString::compare( (*it)->host().upper(), authInfo->host().upper() ) == 0 &&
-              QString::compare( (*it)->share().upper(), authInfo->share().upper() ) == 0) )
-        {
-          authInfo->setUser( (*it)->user() );
-          authInfo->setPassword( (*it)->password() );
-
-          break;
-        }
-        else if ( (QString::compare( (*it)->workgroup().upper(), authInfo->workgroup().upper() ) == 0 &&
-                   QString::compare( (*it)->host().upper(), authInfo->host().upper() ) == 0 &&
-                   QString::compare( (*it)->share().upper(), "*" ) == 0) ||
-                  (QString::compare( (*it)->workgroup().upper(), "*" ) == 0 &&
-                   QString::compare( (*it)->host().upper(), authInfo->host().upper() ) == 0 &&
-                   QString::compare( (*it)->share().upper(), "*" ) == 0) )
-        {
-          authInfo->setUser( (*it)->user() );
-          authInfo->setPassword( (*it)->password() );
-
-          continue;
         }
         else
         {
-          continue;
+          return authInfo;
+        }
+      }
+      else
+      {
+        // Check that the workgroup is correct.
+        if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
+        {
+          if ( QString::compare( map["Workgroup"].upper(), authInfo->workgroup().upper() ) != 0 )
+          {
+            return authInfo;
+          }
+          else
+          {
+            // Everything is OK. Do nothing.
+          }
+        }
+        else
+        {
+          // We cannot check if the workgroup is OK. So, just process
+          // the authentication information we just collected. If it
+          // should be wrong, the user will be asked to provide the
+          // correct data anyway.
         }
       }
     }
-    else if ( !Smb4KSettings::rememberPasswords() )
+
+    // We have authentication information. Put the data into the
+    // Smb4KAuthInfo object.
+    authInfo->setUser( map["Login"] );
+    authInfo->setPassword( map["Password"] );
+  }
+  else
+  {
+    // Either the opening of the wallet failed or the user does
+    // not want to use a wallet.
+    // Let's see if we can get the authentication information from
+    // the temporary list.
+    if ( Smb4KSettings::rememberPasswords() )
+    {
+      if ( !m_auth_list.isEmpty() )
+      {
+        for ( QValueList<Smb4KAuthInfo *>::Iterator it = m_auth_list.begin(); it != m_auth_list.end(); ++it )
+        {
+          if ( !authInfo->share().isEmpty() )
+          {
+            if ( QString::compare( authInfo->host().upper(), (*it)->host().upper() ) == 0 &&
+                 QString::compare( authInfo->share().upper(), (*it)->share().upper() ) == 0 )
+            {
+              if ( QString::compare( authInfo->workgroup().upper(), (*it)->workgroup().upper() ) == 0 ||
+                   (*it)->workgroup().isEmpty() )
+              {
+                // Either the workgroup is OK, or we cannot check if it is OK.
+                // In both cases we process the data at hand.
+                authInfo->setUser( (*it)->user() );
+                authInfo->setPassword( (*it)->password() );
+
+                // Exact match. Stop here.
+                break;
+              }
+              else
+              {
+                continue;
+              }
+            }
+            else if ( QString::compare( authInfo->host().upper(), (*it)->host().upper() ) == 0 )
+            {
+              if ( QString::compare( authInfo->workgroup().upper(), (*it)->workgroup().upper() ) == 0 ||
+                   (*it)->workgroup().isEmpty() )
+              {
+                // Either the workgroup is OK, or we cannot check if it is OK.
+                // In both cases we process the data at hand.
+                authInfo->setUser( (*it)->user() );
+                authInfo->setPassword( (*it)->password() );
+
+                // Because we always prefer the exact match, we keep
+                // on searching.
+                continue;
+              }
+              else
+              {
+                continue;
+              }
+            }
+            else
+            {
+              continue;
+            }
+          }
+          else
+          {
+            if ( QString::compare( authInfo->host().upper(), (*it)->host().upper() ) == 0 )
+            {
+              if ( QString::compare( authInfo->workgroup().upper(), (*it)->workgroup().upper() ) == 0 ||
+                   (*it)->workgroup().isEmpty() )
+              {
+                // Either the workgroup is OK, or we cannot check if it is OK.
+                // In both cases we process the data at hand.
+                authInfo->setUser( (*it)->user() );
+                authInfo->setPassword( (*it)->password() );
+
+                break;
+              }
+              else
+              {
+                continue;
+              }
+            }
+            else
+            {
+              continue;
+            }
+          }
+        }
+      }
+      else
+      {
+        // Do nothing
+      }
+    }
+    else
     {
       if ( m_temp_auth )
       {
@@ -477,70 +657,86 @@ void Smb4KPasswordHandler::writeAuth( Smb4KAuthInfo *authInfo )
 {
   open_close_wallet();
 
-  if ( m_wallet && m_wallet->isOpen() )
+  if ( walletIsOpen() )
   {
-    // Remove old entry:
-    QString key;
-    key.append( authInfo->workgroup().isEmpty() ? "*" : authInfo->workgroup().upper() );
-    key.append( ":" );
-    key.append( authInfo->host().upper() );
-    key.append( ":" );
-    key.append( authInfo->share().isEmpty() ? "*" : authInfo->share().upper() );
+    QMap<QString,QString> map;
+    map["Login"]       = authInfo->user();
+    map["Password"]    = authInfo->password();
 
-    QStringList entries = m_wallet->entryList();
-
-    for ( QStringList::ConstIterator it = entries.begin(); it != entries.end(); ++it )
+    if ( !authInfo->workgroup().isEmpty() )
     {
-      if ( (*it).stripWhiteSpace().startsWith( key ) )
-      {
-        m_wallet->removeEntry( *it );
-
-        break;
-      }
-      else
-      {
-        continue;
-      }
+      map["Workgroup"] = authInfo->workgroup().upper();
     }
 
-    key.append( ":" );
-    key.append( authInfo->user() );
+    if ( !authInfo->share().isEmpty() )
+    {
+      m_wallet->writeMap( "//"+authInfo->host().upper()+"/"+authInfo->share().upper(), map );
+    }
+    else
+    {
+      m_wallet->writeMap( authInfo->host().upper(), map );
+    }
 
-    m_wallet->writePassword( key, authInfo->password() );
     m_wallet->sync();
   }
   else
   {
     if ( Smb4KSettings::rememberPasswords() )
     {
-      bool is_included = false;
+      // Search the temporary list for an entry and modify it. Otherwise
+      // add the new auth info.
+      Smb4KAuthInfo *known = NULL;
 
+//       for ( int i = 0; i < m_auth_list.size(); ++i )
       for ( QValueList<Smb4KAuthInfo *>::Iterator it = m_auth_list.begin(); it != m_auth_list.end(); ++it )
       {
-        if ( (QString::compare( (*it)->workgroup().upper(), authInfo->workgroup().upper() ) == 0 ||
-              QString::compare( (*it)->workgroup().upper(), "*" ) == 0) &&
-             QString::compare( (*it)->host().upper(), authInfo->host().upper() ) == 0 &&
-             QString::compare( (*it)->share().upper(), authInfo->share().upper() ) == 0 )
+        if ( ((*it)->workgroup().isEmpty() ||
+             QString::compare( (*it)->workgroup().upper(),
+             authInfo->workgroup().upper() ) == 0) &&
+             QString::compare( (*it)->host().upper(),
+             authInfo->host().upper() ) == 0 &&
+             ((*it)->share().isEmpty() ||
+             QString::compare( (*it)->share().upper(),
+             authInfo->share().upper() ) == 0) )
         {
-          delete *it;
-          *it = new Smb4KAuthInfo( authInfo->workgroup(), authInfo->host(), authInfo->share(), authInfo->user(), authInfo->password() );
+          known = *it;
 
-          is_included = true;
-
-          break;
+          if ( authInfo->share().isEmpty() == known->share().isEmpty() )
+          {
+            // Exact match. Stop here.
+            break;
+          }
+          else
+          {
+            // Continue until we found the exact match or reached
+            // the end of the list.
+            continue;
+          }
+        }
+        else
+        {
+          continue;
         }
       }
 
-      if ( !is_included )
+      if ( known )
       {
-        m_auth_list.append( new Smb4KAuthInfo( authInfo->workgroup(), authInfo->host(), authInfo->share(), authInfo->user(), authInfo->password() ) );
+        // Overwrite the previous data. The host name and the share name
+        // does not need to be replaced.
+        known->setWorkgroup( authInfo->workgroup() );
+        known->setUser( authInfo->user() );
+        known->setPassword( authInfo->password() );
+      }
+      else
+      {
+        m_auth_list.append( new Smb4KAuthInfo( *authInfo ) );
       }
     }
     else
     {
       if ( !m_temp_auth )
       {
-        m_temp_auth = new Smb4KAuthInfo( authInfo->workgroup(), authInfo->host(), authInfo->share(), authInfo->user(), authInfo->password() );
+        m_temp_auth = new Smb4KAuthInfo( *authInfo );
       }
     }
   }
@@ -562,29 +758,21 @@ Smb4KAuthInfo *Smb4KPasswordHandler::readDefaultAuth( Smb4KAuthInfo *authInfo )
 
   open_close_wallet();
 
-  if ( m_wallet && m_wallet->isOpen() )
+  if ( walletIsOpen() )
   {
-    QStringList full_list = m_wallet->entryList();
-    QStringList default_list = full_list.grep( "DEFAULT" );
+    // Read the default authentication information.
+    QMap<QString,QString> map;
 
-    QString user, pass;
+    m_wallet->readMap( "DEFAULT_LOGIN", map );
 
-    for ( QStringList::ConstIterator it = default_list.begin(); it != default_list.end(); ++it )
+    if ( !map.isEmpty() )
     {
-      if ( (*it).startsWith( "DEFAULT:" ) )
-      {
-        user = (*it).section( ":", 1, 1 ).stripWhiteSpace();
-        m_wallet->readPassword( *it, pass );
-
-        authInfo->setUser( user );
-        authInfo->setPassword( pass );
-
-        break;
-      }
-      else
-      {
-        continue;
-      }
+      authInfo->setUser( map["Login"] );
+      authInfo->setPassword( map["Password"] );
+    }
+    else
+    {
+      // Do nothing
     }
   }
   else
@@ -600,11 +788,14 @@ void Smb4KPasswordHandler::writeDefaultAuth( Smb4KAuthInfo *authInfo )
 {
   open_close_wallet();
 
-  if ( m_wallet && m_wallet->isOpen() )
+  if ( walletIsOpen() )
   {
-    QString key = "DEFAULT:"+authInfo->user();
+    QMap<QString,QString> map;
+    map["Login"]    = authInfo->user();
+    map["Password"] = authInfo->password();
 
-    m_wallet->writePassword( key, authInfo->password() );
+    m_wallet->writeMap( "DEFAULT_LOGIN", map );
+    m_wallet->sync();
   }
 }
 
